# Created by george zhao, who is working for NOKIA, 2010-2012, 2014-2018.												#
# This lib is based on python 3.6.5																						#

MSG_FLAGS_UNSET = 0x0
MSG_COMMAND_CODE_UNSET = 0
MSG_APP_ID_UNSET = 0
MSG_HBH_ID_UNSET = 0x0
MSG_E2E_ID_UNSET = 0x0

MSG_VERSION_BUFF_LEN = 1
MSG_LENGTH_BUFF_LEN = 3
MSG_FLAGS_BUFF_LEN = 1
MSG_COMMAND_CODE_BUFF_LEN = 3
MSG_APPLICATION_ID_BUFF_LEN = 4
MSG_HBH_BUFF_LEN = 4
MSG_E2E_BUFF_LEN = 4
MSG_HEADER_BUFF_LEN = MSG_VERSION_BUFF_LEN + MSG_LENGTH_BUFF_LEN + MSG_FLAGS_BUFF_LEN + MSG_COMMAND_CODE_BUFF_LEN + MSG_APPLICATION_ID_BUFF_LEN + MSG_HBH_BUFF_LEN + MSG_E2E_BUFF_LEN



# Created by george zhao, who is working for NOKIA, 2010-2012, 2014-2018.												#
# This lib is based on python 3.6.5																						#

# Types of AVP, only basic diameter types are supported.																#
from pyDiameter.pyDiaAVPConst import AVP_TYPE_STR
from pyDiameter.pyDiaAVPConst import AVP_TYPE_INT32 
from pyDiameter.pyDiaAVPConst import AVP_TYPE_UINT32 
from pyDiameter.pyDiaAVPConst import AVP_TYPE_INT64
from pyDiameter.pyDiaAVPConst import AVP_TYPE_UINT64 
from pyDiameter.pyDiaAVPConst import AVP_TYPE_FLOAT32
from pyDiameter.pyDiaAVPConst import AVP_TYPE_FLOAT64
from pyDiameter.pyDiaAVPConst import AVP_TYPE_GROUP

# AVP value length in buffer.																							#
AVP_VALUE_INT32_BUFF_LEN = 4
AVP_VALUE_UINT32_BUFF_LEN = 4
AVP_VALUE_INT64_BUFF_LEN = 8
AVP_VALUE_UINT64_BUFF_LEN = 8
AVP_VALUE_FLOAT32_BUFF_LEN = 4
AVP_VALUE_FLOAT64_BUFF_LEN = 8


# Created by george zhao, who is working for NOKIA, 2010-2012, 2014-2018.												#
# This lib is based on python 3.6.5																						#

from random import randint

from pyDiameter.pyDiaAVPTypes import DiaAVPGroup
from pyDiameter.pyDiaAVPFactory import DiaAVPFactory

from pyDiameter.pyDiaMessageConst import MSG_VERSION_BUFF_LEN
from pyDiameter.pyDiaMessageConst import MSG_LENGTH_BUFF_LEN
from pyDiameter.pyDiaMessageConst import MSG_FLAGS_BUFF_LEN
from pyDiameter.pyDiaMessageConst import MSG_COMMAND_CODE_BUFF_LEN
from pyDiameter.pyDiaMessageConst import MSG_APPLICATION_ID_BUFF_LEN
from pyDiameter.pyDiaMessageConst import MSG_HBH_BUFF_LEN
from pyDiameter.pyDiaMessageConst import MSG_E2E_BUFF_LEN
from pyDiameter.pyDiaMessageConst import MSG_HEADER_BUFF_LEN
from pyDiameter.pyDiaMessageConst import MSG_FLAGS_UNSET
from pyDiameter.pyDiaMessageConst import MSG_COMMAND_CODE_UNSET
from pyDiameter.pyDiaMessageConst import MSG_APP_ID_UNSET
from pyDiameter.pyDiaMessageConst import MSG_HBH_ID_UNSET
from pyDiameter.pyDiaMessageConst import MSG_E2E_ID_UNSET


from pyDiameter.pyDiaConst import ZERO_LEN_BYTES
from pyDiameter.pyDiaConst import DEFAULT_BYTES_ORDER

class DiaMessage:
	def __init__(self):
		self.__version = 0x01			# Only Diameter Version 1 is supported by this lib.								#
		self.__flags = MSG_FLAGS_UNSET
		self.__commandCode = MSG_COMMAND_CODE_UNSET
		self.__applicationID = MSG_APP_ID_UNSET
		self.__hbhID = MSG_HBH_ID_UNSET
		self.__e2eID = MSG_E2E_ID_UNSET
		self.__avps = []
		
	# For properties operations																							#
	def getE2EID(self):
		return self.__e2eID
	def setE2EID(self, e2eID):
		self.__e2eID = e2eID
		
	def getHBHID(self):
		return self.__hbhID
	def setHBHID(self, hbhID):
		self.__hbhID = hbhID
		
	def setFlags(self, flags):
		self.__flags = flags
	def getFlags(self):
		return self.__flags
	def setRequestFlag(self):
		self.__flags = self.__flags | 0x80
	def clearRequestFlag(self):
		self.__flags = self.__flags & 0x7F
	def getRequestFlag(self):
		if 1 == (self.__flags & 0x80) >> 7: return True
		return False
	def setProxyableFlag(self):
		self.__flags = self.__flags | 0x40
	def clearProxyableFlag(self):
		self.__flags = self.__flags & 0xBF
	def getProxyableFlag(self):
		if 1 == (self.__flags & 0x40) >> 6: return True
		return False
	def setEBitFlag(self):
		self.__flags = self.__flags | 0x20
	def clearEBitFlag(self):
		self.__flags = self.__flags & 0xDF
	def getEBitFlag(self):
		return (self.__flags & 0x20) >> 5
	def setTBitFlag(self):
		self.__flags = self.__flags | 0x10
	def clearTBitFlag(self):
		self.__flags = self.__flags & 0xEF
	def getTBitFlag(self):
		return (self.__flags & 0x10) >> 4
		
	def setCommandCode(self, cmdCode):
		self.__commandCode = cmdCode
	def getCommandCode(self):
		return self.__commandCode
		
	def setApplicationID(self, appID):
		self.__applicationID = appID
	def getApplicationID(self):
		return self.__applicationID
		
	def generateHBHID(self):
		self.setHBHID(randint(0, 0xffffffff))
	def generateE2EID(self):
		self.setE2EID(randint(0, 0xffffffff))
		
	# getAVPs() is just a function use to debug, will be removed after release#
	def getAVPs(self):
		return self.__avps		
		
		
	# By using DiaAVPPath object, this function can help user to visit AVP directly.									#
	# avpPath parameter should be initialed first by using AVP path language, like 10145/1400[0]->0/493[0]				#
	def getAVPByPath(self, avpPath):
		fakeRoot = DiaAVPGroup()
		fakeRoot.setAVPValue(self.__avps)
		avp = fakeRoot.getAVPByPath(avpPath)
		del fakeRoot
		return avp
		
	# By using DiaAVPPath object, this function can help user to add AVP directly.										#
	# groupAVPPath parameter should be initialed first by using AVP path language, like 10145/1400[0]->0/493[0]			#
	# Note that, groupAVPPath is the father avp which contains the avp you want to add.									#
	# groupAVPPath must be a group type AVP, or you can leave groupAVPPath a blank, avp will be added at root.			#
	def addAVPByPath(self, groupAVPPath, avp, index = -1):
		fakeRoot = DiaAVPGroup()
		fakeRoot.setAVPValue(self.__avps)
		r = fakeRoot.addAVPByPath(groupAVPPath, avp, index)
		del fakeRoot
		return r
		
	# By using DiaAVPPath object, this function can help user to remove AVP directly.									#
	# avpPath parameter should be initialed first by using AVP path language, like 10145/1400[0]->0/493[0]				#
	def removeAVPByPath(self, avpPath):
		fakeRoot = DiaAVPGroup()
		fakeRoot.setAVPValue(self.__avps)
		r = fakeRoot.removeAVPByPath(avpPath)
		del fakeRoot
		return r
			
	# decode tools functions																							#
	# decode tools functions need bytes buff as their parameters, not string.											#		
	# decode an integer value.		
	@classmethod																		#
	def decodeUIntValue(self, intValueBytes):
		assert type(intValueBytes) is bytes
		return int.from_bytes(intValueBytes, DEFAULT_BYTES_ORDER)
		
	# decode only diameter message part.																				#
	def __decodeMessageHeader(self, headerBytesBuff):
		p = 0
		version = self.decodeUIntValue(headerBytesBuff[p:p+MSG_VERSION_BUFF_LEN])
		p += MSG_VERSION_BUFF_LEN
		if self.__version != version:
			raise RuntimeError("Diameter version is not 1 but", version)
		# Length value is ignored.
		p += MSG_LENGTH_BUFF_LEN
		flags = self.decodeUIntValue(headerBytesBuff[p:p+MSG_FLAGS_BUFF_LEN])
		self.setFlags(flags)
		p += MSG_FLAGS_BUFF_LEN
		cmdCode = self.decodeUIntValue(headerBytesBuff[p:p+MSG_COMMAND_CODE_BUFF_LEN])
		self.setCommandCode(cmdCode)
		p += MSG_COMMAND_CODE_BUFF_LEN
		appID = self.decodeUIntValue(headerBytesBuff[p:p+MSG_APPLICATION_ID_BUFF_LEN])
		self.setApplicationID(appID)
		p += MSG_APPLICATION_ID_BUFF_LEN
		hbhID = self.decodeUIntValue(headerBytesBuff[p:p+MSG_HBH_BUFF_LEN])
		self.setHBHID(hbhID)
		p += MSG_HBH_BUFF_LEN
		e2eID = self.decodeUIntValue(headerBytesBuff[p:p+MSG_E2E_BUFF_LEN])
		self.setE2EID(e2eID)
		
	# decode diameter message from a bytes-like buffer to the object.													#
	def decode(self, bytesBuff):
		assert type(bytesBuff) is bytes
		self.__decodeMessageHeader(bytesBuff[0:MSG_HEADER_BUFF_LEN])
		avpsBytesBuff = DiaAVPGroup.cutAVPBytes(bytesBuff[MSG_HEADER_BUFF_LEN:])
		avpFactory =  DiaAVPFactory ()
		for avpBuffsBuff in avpsBytesBuff:
			avp = avpFactory.generateDiaAVPObject(avpBuffsBuff)
			if not None == avp:
				self.__avps.append(avp)
			else:
				raise RuntimeError("Got a Null Pointor when decoding Diameter AVP.")
		
	# encode tools functions																							#
	# encode tools will output bytes, not string.																		#
	# encode an integer value.																							#
	def encodeUIntValue(self, intValue, intValueLength):
		return intValue.to_bytes(intValueLength, DEFAULT_BYTES_ORDER)
		
	# decode only diameter message part.																				#
	def __encodeMessageHeader(self):
		r = ZERO_LEN_BYTES
		r += self.encodeUIntValue(self.__version, MSG_VERSION_BUFF_LEN)
		r += self.encodeUIntValue(len(self), MSG_LENGTH_BUFF_LEN)
		r += self.encodeUIntValue(self.getFlags(), MSG_FLAGS_BUFF_LEN)
		r += self.encodeUIntValue(self.getCommandCode(), MSG_COMMAND_CODE_BUFF_LEN)
		r += self.encodeUIntValue(self.getApplicationID(), MSG_APPLICATION_ID_BUFF_LEN)
		r += self.encodeUIntValue(self.getHBHID(), MSG_HBH_BUFF_LEN)
		r += self.encodeUIntValue(self.getE2EID(), MSG_E2E_BUFF_LEN)
		return r
		
	# encode this object to a bytes-like buffer.																		#
	def encode(self):
		r = ZERO_LEN_BYTES
		r += self.__encodeMessageHeader()
		for avp in self.__avps:
			r += avp.encode()
		return r
	
	# Override the len operator.
	def __len__(self):
		r = MSG_HEADER_BUFF_LEN
		for avp in self.__avps:
			r += avp.getAlignmentLength(len(avp))
		return r


# Created by george zhao, who is working for NOKIA, 2010-2012, 2014-2018.												#
# This lib is based on python 3.6.5																						#

# This class is defined for locating a AVP in a Diameter Message.														#
# When it is used in DiaMessage, it is a path start from the root of the message.										#
# The DiaAVPPath use a path index, defined as: "vendor/avpcode[sameAVPIndex]->vendor/avpCode[sameAVPIndex]->..."		#
# e.g. "10415/517[0]->10415/519[1]->10415/507[0]"																		#
# In this path index, vendor is vendor code for the AVP, avpCode is code of the AVP.									#
# sameAVPIndex means which AVP you want to visit, if there are several AVPs have same vendor and code.					#

import re

class DiaAVPPath:
	def __init__(self):
		self.__pathArray=[]
		
	def __analyzePath(self, path):
		seperator = "->"
		nodes = []
		start = 0
		while True:
			end = path.find(seperator, start)
			if -1 == end:
				break
			else:
				if not start == end:
					nodes.append(path[start:end])
					start = end + len(seperator)
				else:
					break;
		if -1 == end:
			nodes.append(path[start:])
		return nodes
		
	def __analyzeNode(self, node):
		try:
			vendor = int(node[:node.find("/")])
			code = int(node[node.find("/")+1:node.find("[")])
			sai = int(node[node.find("[")+1:node.find("]")])
		except ValueError:
			raise ValueError("For vendor, code and sameAVPIndex, only number is allowed.")
			return None
		if sai < 0: raise ValueError("SameAVPIndex should be a positive value or zero.")
		r = [vendor, code, sai]
		return r
		
	def __analyzePathNodes(self, nodes):
		for node in nodes:
			s = re.match("\d+/\d+\[\d+]$", node)
			if not None == s:
				avpData = self.__analyzeNode(node)
				if not None == avpData:
					self.__pathArray.append(avpData)
			else:
				break;
		
	def setPath(self, path):
		nodes = self.__analyzePath(path)
		self.__analyzePathNodes(nodes)
		
	def getLayersCount(self):
		return len(self.__pathArray)
		
	def getAVPVendor(self, layer):
		return self.__pathArray[layer][0]
		
	def getAVPCode(self, layer):
		return self.__pathArray[layer][1]
	
	def getAVPSameAVPIndex(self, layer):
		return self.__pathArray[layer][2]
		
	def removeTop(self):
		if self.getLayersCount() >= 0:
			del self.__pathArray[0]
		
	def removeTail(self):
		if self.getLayersCount() >= 0:
			del self.__pathArray[len(self.__pathArray)-1]
			
	def getPath(self):
		i = 0
		r = ''
		while i < self.getLayersCount():
			if not i == self.getLayersCount() - 1:
				r += str(self.getAVPVendor(i)) + "/" + str(self.getAVPCode(i)) + "["+str(self.getAVPSameAVPIndex(i)) + "]"+"->"
			else:
				r += str(self.getAVPVendor(i)) + "/" + str(self.getAVPCode(i)) + "["+str(self.getAVPSameAVPIndex(i)) + "]"
			i += 1
		return r


# Created by george zhao, who is working for NOKIA, 2010-2012, 2014-2018.												#
# This lib is based on python 3.6.5																						#

import importlib

from pyDiameter.pyDiaAVPBasicTypesConst import AVP_TYPE_STR
from pyDiameter.pyDiaAVPBasicTypesConst import AVP_TYPE_INT32
from pyDiameter.pyDiaAVPBasicTypesConst import AVP_TYPE_UINT32
from pyDiameter.pyDiaAVPBasicTypesConst import AVP_TYPE_INT64
from pyDiameter.pyDiaAVPBasicTypesConst import AVP_TYPE_UINT64
from pyDiameter.pyDiaAVPBasicTypesConst import AVP_TYPE_FLOAT32
from pyDiameter.pyDiaAVPBasicTypesConst import AVP_TYPE_FLOAT64
from pyDiameter.pyDiaAVPBasicTypesConst import AVP_TYPE_GROUP

# this class is used to generate DiaAVP Object. (An empty object, need call decode)										#
# after get object from this factory, call object decode method to fullfill the object.									#
class DiaAVPFactory():
	def __init__(self):
		self.__typesDict={
			AVP_TYPE_STR: "DiaAVPStr",
			AVP_TYPE_INT32: "DiaAVPInt32",
			AVP_TYPE_UINT32: "DiaAVPUInt32",
			AVP_TYPE_INT64: "DiaAVPInt64",
			AVP_TYPE_UINT64: "DiaAVPUInt64",
			AVP_TYPE_FLOAT32: "DiaAVPFloat32",
			AVP_TYPE_FLOAT64: "DiaAVPFloat64",
			AVP_TYPE_GROUP: "DiaAVPGroup",
		}
		
	def generateDiaAVPObject(self, avpBytes):
		typesModule = importlib.import_module("pyDiameter.pyDiaAVPTypes")
		utAVP = getattr(typesModule, "DiaAVPUnknowType")()
		utAVP.decode(avpBytes)
		type = utAVP.getAVPType()
		if type in self.__typesDict: r = getattr(typesModule, self.__typesDict[type])()
		else: r = None
		if not (None ==  r): r.decode(avpBytes)
		return r
			
# end of DiaAVPFactory.


# Created by george zhao, who is working for NOKIA, 2010-2012, 2014-2018.												#
# This lib is based on python 3.6.5																						#

from abc import ABCMeta
from abc import abstractmethod
from pyDiameter.pyDiaAVPDict import DiaAVPDict

from pyDiameter.pyDiaAVPConst import AVP_VENDOR_UNKNOWN
from pyDiameter.pyDiaAVPConst import AVP_CODE_UNKNOWN
from pyDiameter.pyDiaAVPConst import AVP_NAME_NONE
from pyDiameter.pyDiaAVPConst import AVP_TYPE_UNKNOWN
from pyDiameter.pyDiaAVPConst import AVP_FLAG_CLEAR
from pyDiameter.pyDiaAVPConst import AVP_VALUE_UNSET
from pyDiameter.pyDiaAVPConst import AVP_CODE_BUFF_LEN
from pyDiameter.pyDiaAVPConst import AVP_FLAGS_BUFF_LEN
from pyDiameter.pyDiaAVPConst import AVP_LENGTH_BUFF_LEN
from pyDiameter.pyDiaAVPConst import AVP_VENDOR_BUFF_LEN
from pyDiameter.pyDiaConst import NULL_BYTE
from pyDiameter.pyDiaConst import DEFAULT_BYTES_ORDER
from pyDiameter.pyDiaAVPFactory import DiaAVPFactory


# DiaAVP class is defined for carry Diameter AVP, and its sub-AVPs.														#
class DiaAVP(metaclass=ABCMeta):
	# DiaAVP is an abstract Base Class.

	# init function of DiaAVP.																							#
	def __init__(self):
		self.__avpVendor = AVP_VENDOR_UNKNOWN
		self.__avpCode = AVP_CODE_UNKNOWN
		self.__avpName = AVP_NAME_NONE
		self.__avpType = AVP_TYPE_UNKNOWN
		self.__avpFlags = AVP_FLAG_CLEAR
		self._initAVPValue()

		
	# For property avpName.
	def getAVPName(self):
		vendor = 0
		if self.getAVPFlags():
			vendor = self.getAVPVendor()
		code = self.getAVPCode()
		dict =  DiaAVPDict()
		if dict.hasAVPDef(vendor, code):
			return dict.getAVPDefName(self.getAVPVendor(), self.getAVPCode())
		return AVP_NAME_NONE
		
	# For property avpType, read only.
	def getAVPType(self):
		vendor = 0
		if self.getAVPFlags():
			vendor = self.getAVPVendor()
		code = self.getAVPCode()
		dict =  DiaAVPDict()
		if dict.hasAVPDef(vendor, code):
			return dict.getAVPDefType(self.getAVPVendor(), self.getAVPCode())
		return AVP_TYPE_UNKNOWN
		
	# For property avpCode
	def getAVPCode(self):
		return self.__avpCode
	def setAVPCode(self, avpCode):
		self.__avpCode = avpCode
		return True

	# For property avpFlag
	def getAVPFlags(self):
		return self.__avpFlags
	def setAVPFlags(self, avpFlags):
		self.__avpFlags = avpFlags
		return True
	def setAVPVSFlag(self):
		self.__avpFlags = self.__avpFlags | 0x80
		return True
	def clearAVPVSFlag(self):
		self.__avpFlags = self.__avpFlags & 0x7F
		return True
	def getAVPVSFlag(self):
		if ((self.__avpFlags & 0x80) >> 7) == 1:
			return True
		return False
	def setAVPMandatoryFlag(self):
		self.__avpFlags = self.__avpFlags | 0x40
		return True
	def clearAVPMandatoryFlag(self):
		self.__avpFlags = self.__avpFlags & 0xBF
		return True
	def getAVPMandatoryFlag(self):
		if ((self.__avpFlags & 0x40) >> 6) == 1:
			return True
		return False
	def setAVPProtectedFlag(self):
		self.__avpFlags = self.__avpFlags | 0x20
		return True
	def clearAVPProtectedFlag(self):
		self.__avpFlags = self.__avpFlags & 0xDF
		return True
	def getAVPProtectedFlag(self):
		if ((self.__avpFlags & 0x20) >> 5) == 1:
			return True
		return False
		
	# For property avpVendor.
	def getAVPVendor(self):
		return self.__avpVendor
	def setAVPVendor(self, avpVendor):
		self.__avpVendor = avpVendor
		return True
		
	# For property avpValue
	def getAVPValue(self):
		return self.__avpValue
	def setAVPValue(self, value):
		assert self._validAVPValue(value)
		self.__avpValue = value
		
	# For get AVP header length, read only.
	def getAVPHeaderLength(self):
		l = AVP_CODE_BUFF_LEN + AVP_FLAGS_BUFF_LEN + AVP_LENGTH_BUFF_LEN 
		if self.getAVPVSFlag(): l += AVP_VENDOR_BUFF_LEN
		return l
	
	@classmethod
	def getAlignmentLength(self, len):
		return (len % 4 == 0) and len or (int(len/4) + 1) * 4		
		
	# decode tools functions																							#
	# decode tools functions need bytes buff as their parameters, not string.											#		
	# decode an integer value.		
	@classmethod																		#
	def decodeUIntValue(self, intValueBytes):
		assert type(intValueBytes) is bytes
		return int.from_bytes(intValueBytes, DEFAULT_BYTES_ORDER)
	
	# decode AVP code function																							#	
	def decodeAVPCode(self, codeBytes):
		assert (AVP_CODE_BUFF_LEN == len(codeBytes))
		return self.decodeUIntValue(codeBytes)
		
	# decode AVP Flags function.																						#
	def decodeAVPFlags(self, flagsBytes):
		assert (AVP_FLAGS_BUFF_LEN == len(flagsBytes))
		return self.decodeUIntValue(flagsBytes)
	
	# decode AVP length function.																						#
	def decodeAVPLength(self, lengthBytes):	
		assert (AVP_LENGTH_BUFF_LEN == len(lengthBytes))	
		return self.decodeUIntValue(lengthBytes)	
			
	# decode AVP vendor function.																						#
	def decodeAVPVendor(self, vendorBytes):
		assert (AVP_VENDOR_BUFF_LEN == len(vendorBytes))
		return self.decodeUIntValue(vendorBytes)
	
	# encode tools functions																							#
	# encode tools will output bytes, not string.																		#
	# encode an integer value.																							#
	def encodeUIntValue(self, intValue, intValueLength):
		return intValue.to_bytes(intValueLength, DEFAULT_BYTES_ORDER)
		
	# encode AVP code function																							#	
	def encodeAVPCode(self, avpCode):
		assert type(avpCode) is int
		return self.encodeUIntValue(avpCode, AVP_CODE_BUFF_LEN)
		
	# encode AVP vendor function																						#	
	def encodeAVPFlags(self, avpFlags):
		assert type(avpFlags) is int
		return self.encodeUIntValue(avpFlags, AVP_FLAGS_BUFF_LEN)
		
	# encode AVP length function																						#	
	def encodeAVPLength(self, avpLength):
		assert type(avpLength) is int
		return self.encodeUIntValue(avpLength, AVP_LENGTH_BUFF_LEN)		
		
	# encode AVP vendor function																						#	
	def encodeAVPVendor(self, avpVendor):
		assert type(avpVendor) is int
		return self.encodeUIntValue(avpVendor, AVP_VENDOR_BUFF_LEN)		
	
	# decode is for decoding a binary bytes to DiaAVP object.															#
	def decode(self, avpBuffBytes):
		p = 0
		self.setAVPCode(self.decodeAVPCode(avpBuffBytes[p:p+AVP_CODE_BUFF_LEN]))
		p += AVP_CODE_BUFF_LEN
		self.setAVPFlags(self.decodeAVPFlags(avpBuffBytes[p:p+AVP_FLAGS_BUFF_LEN]))
		p += AVP_FLAGS_BUFF_LEN
		tmpLen = (self.decodeAVPLength(avpBuffBytes[p:p+AVP_LENGTH_BUFF_LEN]))
		p += AVP_LENGTH_BUFF_LEN
		if self.getAVPVSFlag():
			self.setAVPVendor(self.decodeAVPVendor(avpBuffBytes[p:p+AVP_VENDOR_BUFF_LEN]))
			p += AVP_VENDOR_BUFF_LEN
		valueLen = tmpLen - p
		assert tmpLen <= len(avpBuffBytes), "Buffer length is smaller than length indicated in AVP."
		self.setAVPValue(self._decodeAVPValue(avpBuffBytes[p:p+valueLen]))
		return True
		
	# encode is for encoding a DiaAVP object to a binary bytes.															#
	def encode(self):
		retValue = self.encodeAVPCode(self.getAVPCode())
		retValue += self.encodeAVPFlags(self.getAVPFlags())
		retValue += self.encodeAVPLength(len(self))
		if self.getAVPVSFlag(): retValue += self.encodeAVPVendor(self.getAVPVendor())
		encodedValue = self._encodeAVPValue(self.getAVPValue())
		l = len(encodedValue)
		retValue += encodedValue + NULL_BYTE * (self.getAlignmentLength(l) - l)
		return retValue
		
	# Children class will override following function.
	@abstractmethod
	def _decodeAVPValue(self, valueBytes):
		raise NotImplementedError	
	@abstractmethod
	def _encodeAVPValue(self, valueBytes):
		raise NotImplementedError
	@abstractmethod
	def _validAVPValue(self, value):
		raise NotImplementedError
	@abstractmethod
	def _initAVPValue(self):
		raise NotImplementedError
	# For get AVP length
	@abstractmethod
	def __len__(self):
		raise NotImplementedError
	# For get sub-AVP in this AVP object.
	@abstractmethod
	def __getitem__(self, index):
		raise NotImplementedError
		
# end of DiaAVP class.




# Created by george zhao, who is working for NOKIA, 2010-2012, 2014-2018.												#
# This lib is based on python 3.6.5																						#

from pyDiameter.pyDiaAVPConst import AVP_ARRAY_DEFAULT
from pyDiameter.pyDiaAVPConst import NAME_KEY, TYPE_KEY
from pyDiameter.pyDiaAVPConst import AVP_TYPE_UNKNOWN
from pyDiameter.pyDiaAVPConst import AVP_NAME_NONE

# DiaAVPDict is used to clarify the class DiaAVPDict, which is contained a default dictionary,							#
# in this default dictionary, almost diameter AVPs were stored, including their name, code, verdor and type.			#		
# But anyway, this default dictionary could not contains all AVP registered in all specifications all over the world.	#
# it only contains a popular AVPs' definition, e.g. base diameter, and 3gpp (10415).									#
# The DiaAVPDict class provide methods to add your own AVP definitions, when you run your app based on it.				#

class DiaAVPDict():
	__instance=None

	def __new__(cls, *args, **kwargs):
		if cls.__instance is None:
			cls.__instance = super().__new__(cls)
		return cls.__instance
		
	def __init__(self):
		self.__avpDict = AVP_ARRAY_DEFAULT
		
	# Create a new vendor in default dictionary.																		#
	def __createAVPVendor(self, avpVendor):
		self.__avpDict[avpVendor]={}
		
	# Create a new AVP code in a vendor scope in default dictionary.													#
	def __createAVPCode(self ,avpVendor, avpCode):
		self.__avpDict[avpVendor][avpCode]={}
		
	# Judge whether Vendor is defined in default dictionary.															#
	def hasVendorDef(self, avpVendor):
		if avpVendor in self.__avpDict:
			return True
		return False
		
	# Judge whether AVP is exist in default dictionary. If yes, True will be returned, and False for not.				#
	def hasAVPDef(self, avpVendor, avpCode):
		if avpVendor in self.__avpDict:
			if avpCode in self.__avpDict[avpVendor]:
				return True
		return False
		
	# Add AVP definition if relavant definition is not exist in default dictionary.										#
	# If AVP is already exist in default dictionary, this function will raise an KeyError, 								#
	# with comment "Key is already exist, please use function updateAVPDef() to update the AVP definition"				#
	# avpType should be one of CONSTs, TYPE_STR, TYPE_INT32, TYPE_INT64, TYPE_GROUP										#
	def addAVPDef(self, avpVendor, avpCode, avpName, avpType):
		if self.hasAVPDef(avpVendor, avpCode):
			raise KeyError("AVP vendor and code is already exist, please use function updateAVPDef() to update the AVP definition.")
			return False
		if not self.hasVendorDef(avpVendor):
			self.__createAVPVendor(avpVendor)
		self.__createAVPCode(avpVendor, avpCode)
		self.__avpDict[avpVendor][avpCode][NAME_KEY] = avpName
		self.__avpDict[avpVendor][avpCode][TYPE_KEY] = avpType
		return True
		
	# Update AVP definition according to parameters.																	#
	# If AVP is not exist in default dictionary, this function will raise an KeyError,									#
	# with comment "Key is not exist, please use function addAVPDef() to add new AVP definition."						#
	# avpType should be one of CONSTs, TYPE_STR, TYPE_INT32, TYPE_INT64, TYPE_GROUP										#
	def updateAVPDef(self, avpVendor, avpCode, avpName, avpType):
		if not self.hasVendorDef(avpVendor):
			raise KeyError("AVP vendor is not exist, please use function addAVPDef() to add new AVP definition.")
			return False			
		if not self.hasAVPDef(avpVendor, avpCode):
			raise KeyError("AVP code is not exist, please use function addAVPDef() to add new AVP definition.")
			return False
		self.__avpDict[avpVendor][avpCode][NAME_KEY] = avpName
		self.__avpDict[avpVendor][avpCode][TYPE_KEY] = avpType
		return True		
	
	# In this version, delete the AVP definition is not supported currently.											#
	def deleteAVPDef(self, avpVendor, avpCode):
		raise NotImplementedError("The function deleteAVPDef() is not implementated currently.")

	# get AVP name by using vendor and code.
	def getAVPDefName(self, avpVendor, avpCode):
		if self.hasAVPDef(avpVendor, avpCode):
			return self.__avpDict[avpVendor][avpCode][NAME_KEY]
		return AVP_NAME_NONE
	
	# get AVP type by using vendor and code.
	def getAVPDefType(self, avpVendor, avpCode):
		if self.hasAVPDef(avpVendor, avpCode):
			return self.__avpDict[avpVendor][avpCode][TYPE_KEY]
		return AVP_TYPE_UNKNOWN

# end of class DiaAVPDict
